@@ -44,8 +44,9 @@ info =
 main :: IO ()
 main = do
     Opts{..} <- cmdArgs opts
-    let conf = Config { confIndent  = indent
-                      , confCompare = if sort then compare else mempty
+    let conf = Config { confIndent    = Spaces indent
+                      , confCompare   = if sort then compare else mempty
+                      , confNumFormat = Generic
                       }
         enc = if compact then encode else encodePretty' conf
     interact $ unlines . map enc . values
